/*****************************************************************************
*
* Description: IDC Tracer
* Author: pva <pva_co@ukr.net>
*
* Copyright (c) 2005, pva All rights reserved.
*
* $Log$
*****************************************************************************/
//
// TODO:
//	collect all branches to list
//	mark visited locations by marker - may be in comments
//	if loop detected -> switch to next branch
//
//	if call - allow user assign name to location
//
#include <idc.idc>
#include "service.idc"

#define	TRACE_FILE	"output.log"

static main(void)
{
	auto	_GLOBALS_;		// current state storage between runs
					// such parameters as:
					//  - silent mode
					//  - jxx mode

	auto	_STACK_,		// stack storage
		_STACK_POS_;

	auto	flog;			// Log file

	auto	r_eax,			// CPU register set
		r_ebx,
		r_ecx,
		r_edx,
		r_esi,
		r_edi,
		r_ebp;

	auto	lStartPos,		// Start analysis possition
		sStartName;		// and it's name
	auto	lCurrentPos,		// current position
		lPrevPos,		// previouse position used to detect jmp $ deadloops
		sCurrentString,		// disasm full string
		sCurrentMnem,		// instruction mnemonic
		sCurrentFlags,		// IDA flags for current line
		sCurrentOp1,		// instruvtion operands
		sCurrentOp2;		// 

	auto	bLastJxx;		// jx/jnx detection flag
	auto	silent_;		// silent mode between pushfw/popfw pair

	auto	addr_, offs_, size_, str_;
	auto	item_;
	auto 	i, j;

	//
	// Create _GLOBALS_
	//
	_GLOBALS_	= GetArrayId("_GLOBALS_");
	if( _GLOBALS_ == -1 )
	{
		_GLOBALS_	= CreateArray("_GLOBALS_");
		if( _GLOBALS_ == -1 )
		{
			Message("[ERROR]: Can't create globals.\n");
			return -1;
		}
		silent_  = 0;
		bLastJxx = 0;
	}
	else
	{
		silent_	 = GetArrayElement(AR_LONG, _GLOBALS_, 1);
		bLastJxx = GetArrayElement(AR_LONG, _GLOBALS_, 2);
	}

	//
	// Prepare stack
	//
	_STACK_	= GetArrayId("_STACK_");
	if( _STACK_ == -1 )
	{
		_STACK_		= CreateArray("_STACK_");
		if( _STACK_ == -1 )
		{
			Message("[ERROR]: Can't create stack.\n");
			return -1;
		}
	}
	else
	{
		_STACK_POS_	= GetArrayElement(AR_LONG, _STACK_, -1);
	}

	_STACK_POS_	= AskLong(_STACK_POS_, "Start stack size");
	if( _STACK_POS_ == -1 )
	{
		_STACK_POS_	= 0;
	}

	//
	// Prepare trace log
	//
	flog	= fopen(TRACE_FILE, "a");
	if( flog == 0 )
	{
		Message("[ERROR]: Can't open log file.\n");
		return -1;
	}


	//
	// Start MAIN loop
	//
	lStartPos	= ScreenEA();
	sStartName	= Name(lStartPos);
	lCurrentPos	= lStartPos;
	lPrevPos	= BADADDR;

	Message("[info]: Start at: %08x (%s)\n", lStartPos, sStartName);
	do
	{
		MakeUnkn(lCurrentPos, 1);
		Jump(lCurrentPos);
		MakeCode(lCurrentPos);
		sCurrentString	= GetDisasm(lCurrentPos);
		sCurrentMnem	= GetMnem(lCurrentPos);
		sCurrentOp1	= makeOperand(lCurrentPos, 0);
		sCurrentOp2	= makeOperand(lCurrentPos, 1);

		//
		// Check for jmp $ or same
		//
		if( lPrevPos != lCurrentPos )
		{
			//
			// Log full trace info to log
			//
			fprintf(flog, "loc_%x:\n", lCurrentPos);
			if( hasName(GetFlags(lCurrentPos)) ) fprintf(flog, "%s:\n", Name(lCurrentPos));
			lPrevPos = lCurrentPos;
		}
		else
		{
			//
			// Stop if deadloop detected
			//
			Message("[info]: finished on deadloop at: %08x (%s)\n", lCurrentPos, Name(lCurrentPos));
			break;
		}

		//
		// Reset LastJxx on cmp or same
		//
		if( !silent_ )
		{
			if( (strstr(sCurrentMnem, "and") != -1)
			 || (strstr(sCurrentMnem, "cmp") != -1)
			 || (strstr(sCurrentMnem, "or") != -1)
			 || (strstr(sCurrentMnem, "sub") != -1)
			 || (strstr(sCurrentMnem, "test") != -1)
			 || (strstr(sCurrentMnem, "xor") != -1)
			) bLastJxx = 0;
		}

		//
		// Virtual CPU below
		// Each function that can change state of VM has own branch
		// Common behaviour is
		//  - log current string
		//  - change state
		//  - set lCurrentPos to new location or to BADADDR if break need
		//
		if( strstr(sCurrentMnem, "mov") != -1 )
		{
			LogCmd();
			lCurrentPos	= getNextAddr(lCurrentPos);

			op2num(sCurrentOp2, item_, 1);
			num2reg(sCurrentOp1, item_);
		}
		else
		if( strstr(sCurrentMnem, "and") != -1 )
		{
			LogCmd();
			lCurrentPos	= getNextAddr(lCurrentPos);

			op2num(sCurrentOp2, item_, 1);
			reg_num(sCurrentOp1, item_, &);
		}
		else
		if( strstr(sCurrentMnem, "xor") != -1 )
		{
			LogCmd();
			lCurrentPos	= getNextAddr(lCurrentPos);

			op2num(sCurrentOp2, item_, 1);
			reg_num(sCurrentOp1, item_, ^);
		}
		else
		if( strstr(sCurrentMnem, "or") != -1 )
		{
			LogCmd();
			lCurrentPos	= getNextAddr(lCurrentPos);

			op2num(sCurrentOp2, item_, 1);
			reg_num(sCurrentOp1, item_, |);
		}
		else
		if( (strstr(sCurrentMnem, "shl") != -1) || (strstr(sCurrentMnem, "shr") != -1) )
		{
			LogCmd();
			lCurrentPos	= getNextAddr(lCurrentPos);

			size_	= makeNumber(sCurrentOp2);
			op2num(sCurrentOp1, item_, 0);

			if( strstr(sCurrentMnem, "shl") != -1 )	item_ = item_ << size_;	else
			if( strstr(sCurrentMnem, "shr") != -1 )	item_ = item_ >> size_;	else
				lCurrentPos = BADADDR;

			num2reg(sCurrentOp1, item_);
		}
		else
		if( strstr(sCurrentMnem, "jmp") != -1 )
		{
			if( strstr(sCurrentOp1, "ds:") == -1 )
			{
				lCurrentPos	= makeAddr(sCurrentOp1);
				if( (lCurrentPos == BADADDR) && (GetOpType(ScreenEA(), 0) == 1) )
				{
					reg2num(sCurrentOp1, lCurrentPos);
				}

				Message("[info]: jmp ");
				if( lCurrentPos == BADADDR ) Message("BAD");
				else			     Message(" Reg = %08x", lCurrentPos);
			}
			else
			{
				Message("[info]: System function");
				lCurrentPos	= getNextAddr(lCurrentPos);
			}
			Message("\n");
		}
		else
		if( (strstr(sCurrentMnem, "j") == 0) && (Byte(lCurrentPos) != 0xe3) ) // skip j[e]cxz as usual cmd
		{
			LogCmd();
			if( bLastJxx )
			{
				i = Byte(lCurrentPos);
				if( i == 0x0f )	i = Byte(lCurrentPos + 1) - 0x10;	// convert to 7x
				if( (bLastJxx ^ i) == 1 )
				{
					Log("; JXX detected\n");
					Message("[info]: JXX jump detected");
					lCurrentPos	= makeAddr(sCurrentOp1);
					if( (lCurrentPos == BADADDR) && (GetOpType(ScreenEA(), 0) == 1) )
					{
						reg2num(sCurrentOp1, lCurrentPos);
					}
					if( lCurrentPos == BADADDR ) Message("BAD");
					else			     Message(" Reg = %08x", lCurrentPos);
					Message("\n");
				}
				else
				{
					lCurrentPos	= getNextAddr(lCurrentPos);
					bLastJxx = i;
				}
			}

			if( !bLastJxx )
			{
				bLastJxx = Byte(lCurrentPos);
				if( bLastJxx == 0x0f )	bLastJxx = Byte(lCurrentPos + 1) - 0x10;
				lCurrentPos	= getNextAddr(lCurrentPos);
			}
		}
		else
		if( strstr(sCurrentMnem, "call") != -1 )
		{
			Message("[info]: %s\n", sCurrentString);
			LogCmd();

			if( strstr(sCurrentOp1, "ds:") != -1 )
			{
				lCurrentPos	= getNextAddr(lCurrentPos);
				if( strstr(sCurrentOp1, "KeDelayExecutionThread") != -1 )	_STACK_POS_ = _STACK_POS_ - 0x0C;	else
				if( strstr(sCurrentOp1, "KeReleaseMutex") != -1 )		_STACK_POS_ = _STACK_POS_ - 0x08;	else
				if( strstr(sCurrentOp1, "KeWaitForSingleObject") != -1 )	_STACK_POS_ = _STACK_POS_ - 0x14;	else
				if( strstr(sCurrentOp1, "MmMapLockedPages") != -1 )		_STACK_POS_ = _STACK_POS_ - 0x08;	else
				if( strstr(sCurrentOp1, "IofCompleteRequest") != -1 )		_STACK_POS_ = _STACK_POS_ - 0x08;	else
				if( strstr(sCurrentOp1, "ExAllocatePool") != -1 )		_STACK_POS_ = _STACK_POS_ - 0x08;	else
				if( strstr(sCurrentOp1, "ExFreePool") != -1 )			_STACK_POS_ = _STACK_POS_ - 0x04;	else
				if( strstr(sCurrentOp1, "IoGetDeviceObjectPointer") != -1 )	_STACK_POS_ = _STACK_POS_ - 0x10;	else
				if( strstr(sCurrentOp1, "RtlInitUnicodeString") != -1 )		_STACK_POS_ = _STACK_POS_ - 0x08;	else
				if( strstr(sCurrentOp1, "ObfDereferenceObject") != -1 )		_STACK_POS_ = _STACK_POS_ - 0x04;	else
				if( strstr(sCurrentOp1, "PsTerminateSystemThread") != -1 )	_STACK_POS_ = _STACK_POS_ - 0x04;	else
				if( strstr(sCurrentOp1, "ZwClose") != -1 )			_STACK_POS_ = _STACK_POS_ - 0x04;	else
				if( strstr(sCurrentOp1, "PsCreateSystemThread") != -1 )		_STACK_POS_ = _STACK_POS_ - 0x1C;	else
				if( strstr(sCurrentOp1, "ObReferenceObjectByHandle") != -1 )	_STACK_POS_ = _STACK_POS_ - 0x18;	else
				if( strstr(sCurrentOp1, "KeInitializeMutex") != -1 )		_STACK_POS_ = _STACK_POS_ - 0x08;	else
				if( strstr(sCurrentOp1, "MmMapIoSpace") != -1 )			_STACK_POS_ = _STACK_POS_ - 0x0C;	else
				if( strstr(sCurrentOp1, "MmUnmapIoSpace") != -1 )		_STACK_POS_ = _STACK_POS_ - 0x08;	else
				if( strstr(sCurrentOp1, "ZwOpenKey") != -1 )			_STACK_POS_ = _STACK_POS_ - 0x0C;	else
				if( strstr(sCurrentOp1, "RtlInitAnsiString") != -1 )		_STACK_POS_ = _STACK_POS_ - 0x08;	else
				if( strstr(sCurrentOp1, "RtlAnsiStringToUnicodeString") != -1 )	_STACK_POS_ = _STACK_POS_ - 0x0C;	else
				if( strstr(sCurrentOp1, "RtlFreeUnicodeString") != -1 )		_STACK_POS_ = _STACK_POS_ - 0x04;	else
				if( strstr(sCurrentOp1, "ZwQueryValueKey") != -1 )		_STACK_POS_ = _STACK_POS_ - 0x18;	else
				if( strstr(sCurrentOp1, "ZwSetValueKey") != -1 )		_STACK_POS_ = _STACK_POS_ - 0x18;	else
				if( strstr(sCurrentOp1, "strchr") != -1 )			_STACK_POS_ = _STACK_POS_ - 0x08;	else
				if( strstr(sCurrentOp1, "sprintf") != -1 )			_STACK_POS_ = _STACK_POS_ - 0x00;	else
				if( strstr(sCurrentOp1, "RtlUnicodeStringToInteger") != -1 )	_STACK_POS_ = _STACK_POS_ - 0x0C;	else
				if( strstr(sCurrentOp1, "ZwEnumerateKey") != -1 )		_STACK_POS_ = _STACK_POS_ - 0x18;	else
				if( strstr(sCurrentOp1, "ZwQueryKey") != -1 )			_STACK_POS_ = _STACK_POS_ - 0x14;	else
					lCurrentPos	= BADADDR;
			}
			else
			if( strstr(sCurrentOp1, "_") == 0 )
			{
				lCurrentPos	= BADADDR;
			}
			else
			{
				lCurrentPos	= makeAddr(sCurrentOp1);
				if( lCurrentPos != BADADDR )
				{
					if( !SetArrayLong(_STACK_, _STACK_POS_, getNextAddr(ScreenEA())) )
					{
						lCurrentPos	= BADADDR;
					}
					else
					{
						_STACK_POS_	= _STACK_POS_ + 4; // 32-bit addr
					}
				}
			}
		}
		else
		if( strstr(sCurrentMnem, "ret") != -1 )
		{
			Message("[info]: %s\n", sCurrentString);
			LogCmd();

			if( _STACK_POS_ > 3 )
			{
				_STACK_POS_	= _STACK_POS_ - 4;
				lCurrentPos	= makeAddr(GetArrayElement(AR_LONG, _STACK_, _STACK_POS_));
				Message("[stack]: %08x => %08x\n", _STACK_POS_, lCurrentPos);

				if( !lCurrentPos )
				{
					lCurrentPos	= BADADDR;
				}
				else
				{
					ExtLinA(ScreenEA(), 0, form("[info]: 0x%08x", lCurrentPos));
					DelExtLnA(ScreenEA(), 1);
				}
			}
			else
			{
				lCurrentPos	= BADADDR;
			}
		}
		else
		if( (strstr(sCurrentMnem, "add") != -1) || (strstr(sCurrentMnem, "sub") != -1) )
		{
			LogCmd();
			lCurrentPos	= getNextAddr(lCurrentPos);

			if( strstr(sCurrentOp1, "dword") != -1 ) size_	= 4;	else
			if( strstr(sCurrentOp1, "word" ) != -1 ) size_	= 2;	else
			if( strstr(sCurrentOp1, "byte" ) != -1 ) size_	= 1;	else
				size_	= 4;

			offs_	= BADADDR; // skip updating
			if( strstr(sCurrentOp1, "[esp") != -1 )
			{
				addr_	= strstr(sCurrentOp1, "+");
				if( addr_ == -1 )
				{
					if( strstr(sCurrentOp1, "[esp]") == -1 )
						lCurrentPos	= BADADDR;
					else
						offs_	= 0;
				}
				else
				{
					addr_	= addr_ + 1; // '+ ' length
					offs_	= strstr(sCurrentOp1, "]");
					str_	= substr(sCurrentOp1, addr_, offs_);
					offs_	= makeNumber(str_);
				}
			}
			else
			if( (strstr(sCurrentOp1, "esp") != -1) && (strlen(sCurrentOp1) == 3) )
			{
				offs_		= makeNumber(sCurrentOp2);
				if( offs_ != 0 )
				{
					if( strstr(sCurrentMnem, "sub") != -1 )
						_STACK_POS_	= _STACK_POS_ + offs_;
					else
						_STACK_POS_	= _STACK_POS_ - offs_;
				}
				else
				{
					lCurrentPos	= BADADDR;
				}
				offs_	= BADADDR; // skip updating
			}
			else
			if( GetOpType(ScreenEA(), 0) == 1 )	// first operand - register
			{
				offs_	= makeNumber(sCurrentOp2);
				if( strstr(sCurrentMnem, "sub") != -1 )
				{
					offs_	= -offs_;
				}

				reg_num(sCurrentOp1, offs_, +); else lCurrentPos = BADADDR;
				offs_	= BADADDR; // skip updating
			}

			//
			// add	[esp], reg
			// sub	[esp], reg
			//
			if( (lCurrentPos != BADADDR) && (offs_ != BADADDR) )
			{
				item_	= GetArrayElement(AR_LONG, _STACK_, _STACK_POS_ - offs_ - size_);
				op2num(sCurrentOp2, i, 1);

				if( strstr(sCurrentMnem, "sub") != -1 )	item_	= item_ - i;
				else					item_	= item_ + i;

				if( !SetArrayLong(_STACK_, _STACK_POS_ - offs_ - size_, item_) )
				{
					lCurrentPos	= BADADDR;
				}

				Message("[stack]: %08x => %08x\n", _STACK_POS_ - offs_ - size_, item_);
			}
		}
		else
		if( strstr(sCurrentMnem, "pushfw") != -1 )
		{
			silent_		= 1;

			Message("[info]: %s\n", sCurrentString);
			LogCmd();

			lCurrentPos	= getNextAddr(lCurrentPos);
			_STACK_POS_	= _STACK_POS_ + 4; // 32-bit addr
		}
		else
		if( strstr(sCurrentMnem, "popfw") != -1 )
		{
			Message("[info]: %s\n", sCurrentString);
			LogCmd();

			lCurrentPos	= getNextAddr(lCurrentPos);
			_STACK_POS_	= _STACK_POS_ - 4; // 32-bit addr

			silent_		= 0;
		}
		else
		if( (strstr(sCurrentMnem, "push") != -1) && (strlen(sCurrentMnem) == 4) )
		{
			Message("[info]: %s\n", sCurrentString);
			LogCmd();

			if( strstr(sCurrentOp1, "dword") != -1 ) size_	= 4;	else
			if( strstr(sCurrentOp1, "word" ) != -1 ) size_	= 2;	else
			if( strstr(sCurrentOp1, "byte" ) != -1 ) size_	= 1;	else
				size_	= 4;

			lCurrentPos	= getNextAddr(lCurrentPos);
			addr_		= makeAddr(sCurrentOp1);
			if( size_ == 4 )
			{
				if( !SetArrayLong(_STACK_, _STACK_POS_, addr_) )
				{
					lCurrentPos	= BADADDR;
				}
				else
				{
					_STACK_POS_	= _STACK_POS_ + size_; // 32-bit addr
					Message("[stack]: position = %08x\n", _STACK_POS_);
				}
			}
		}
		else
		if( (strstr(sCurrentMnem, "pop") != -1) && (strlen(sCurrentMnem) == 3) )
		{
			Message("[info]: %s\n", sCurrentString);
			LogCmd();

			if( strstr(sCurrentOp1, "dword") != -1) size_	= 4;	else
			if( strstr(sCurrentOp1, "word") != -1 ) size_	= 2;	else
			if( strstr(sCurrentOp1, "byte") != -1 ) size_	= 1;	else
				size_	= 4;

			if( size_ == 4)
			{
				_STACK_POS_	= _STACK_POS_ - size_; // 32-bit addr
				Message("[stack]: position = %08x\n", _STACK_POS_);
			}
			lCurrentPos	= getNextAddr(lCurrentPos);
		}
		else
		{
			LogCmd();
			lCurrentPos	= getNextAddr(lCurrentPos);
		}

		//
		// Prepare disasm for next step
		//
		if( lCurrentPos != BADADDR )
		{
			sCurrentFlags	= GetFlags(lCurrentPos);
			if( !isCode(sCurrentFlags) )
			{
				if( !MakeCode(lCurrentPos) )
				{
					lCurrentPos	= BADADDR;
				}
			}
		}

	} while(lCurrentPos != BADADDR);

	//
	// Dump, Stack walk & finish
	//
	j = 0;
	ExtLinA(ScreenEA(), j, form("EAX=%08x, EBX=%08x, ECX=%08x, EDX=%08x", r_eax, r_ebx, r_ecx, r_edx));	++j;
	ExtLinA(ScreenEA(), j, form("ESI=%08x, EDI=%08x, EBP=%08x", r_esi, r_edi, r_ebp));	++j;
	SetArrayLong(_STACK_, -1, _STACK_POS_);
	for(i = _STACK_POS_ - 4; i >= 0; i = i - 4)
	{
		ExtLinA(ScreenEA(), j, form("[stack]: %08x => 0x%08x", i, GetArrayElement(AR_LONG, _STACK_, i)));
		++j;
	}
	DelExtLnA(ScreenEA(), j);

	fclose(flog);
//	DeleteArray(_STACK_);
}
